
-------------------------------------------------------------------------------
-- Optimization
-------------------------------------------------------------------------------

-- GLOBALS: CreateColor, CreateFrame, CreateFramePool, CreateFromMixins, GameTooltip, GameTooltip_SetDefaultAnchor, Mixin, table, UIParent

local MouseIsOver = MouseIsOver
local string = string
local tonumber = tonumber
local type = type
local UNKNOWN = UNKNOWN
local unpack = unpack

-------------------------------------------------------------------------------
-- Namespace
-------------------------------------------------------------------------------

local MAJOR, MINOR = "SixDisplay-1.0", 1
if not LibStub then error(MAJOR .. " requires LibStub.") end
local L = LibStub:NewLibrary(MAJOR, MINOR)
if not L then return end

local AceEvent = LibStub:GetLibrary("AceEvent-3.0", true)
if not AceEvent then return end
AceEvent:Embed(L)

local AceTimer = LibStub:GetLibrary("AceTimer-3.0", true)
if not AceTimer then return end
AceTimer:Embed(L)

local LibDataBroker = LibStub("LibDataBroker-1.1")
local LibSmoothStatusBar = LibStub("LibSmoothStatusBar-1.0")

local CONFIG_DEFAULTS = {
	glyphFont = GameFontNormal,
	glyphs = {},
	textColor = NORMAL_FONT_COLOR,
	textFont = GameFontNormal,
}

-------------------------------------------------------------------------------
-- SixDisplay
-------------------------------------------------------------------------------

local SixDisplayMixin = {}

function SixDisplayMixin.GetTooltipAnchor(frame)
	local x, y = frame:GetCenter()
	if not x or not y then return "TOPLEFT", "BOTTOMLEFT" end
	local h = (x > UIParent:GetWidth() * 2 / 3) and "RIGHT" or (x < UIParent:GetWidth() / 3) and "LEFT" or ""
	return (y > UIParent:GetHeight() / 2) and ("TOP" .. h) or ("BOTTOM" .. h), frame, (y > UIParent:GetHeight() / 2) and "BOTTOM" .. h or "TOP" .. h
end

function SixDisplayMixin.GetZoomedTexCoords(zoomPercent)
	local zoomFactor = zoomPercent / 100
	local quarterZoom = 0.25 * zoomFactor
	return {quarterZoom, 1 - quarterZoom, quarterZoom, 1 - quarterZoom}
end

function SixDisplayMixin.ShouldFillBar(barData)
	return SixDisplayMixin.ShouldResetBar(barData) and barData.value.current < 100
end

function SixDisplayMixin.ShouldResetBar(barData)
	return (barData.level.current ~= barData.level.next or barData.value.current > barData.value.next) and barData.value.current ~= 0
end

function SixDisplayMixin.ShouldSkipAnimation(barData)
	return barData.label.current ~= barData.label.next
end

function SixDisplayMixin:OnAttributeChanged(_, _, key)
	local method = self.KeyToMethod[key]

	if method then
		method(self)
	end

	if MouseIsOver(self) then
		self:OnScript("OnEnter", self)
	end
end

function SixDisplayMixin:OnDataObjectCreated(_, name, dataObject)
	if name == self.config.name then
		self.dataObject = dataObject
		self:SetDataObject()
		self:Update()
	end
end

function SixDisplayMixin:OnScript(script, ...)
	local tooltip = self.dataObject.tooltip or GameTooltip

	if script == "OnEnter" then
		if self.dataObject.type == "progress bar" then
			self:SetBorderColor(self:GetBarColor())
		end

		if self.dataObject.OnEnter then
			self.dataObject.OnEnter(self, ...)
		else
			if self.config.group == "SixUI_ProgressGroup" then
				GameTooltip_SetDefaultAnchor(GameTooltip, self)
			else
				tooltip:SetOwner(self, "ANCHOR_NONE")
				tooltip:SetPoint(self.GetTooltipAnchor(self))
			end

			if self.dataObject.OnTooltipShow then
				self.dataObject.OnTooltipShow(tooltip)
			else
				tooltip:AddLine(self.dataObject.label or self.config.name)
			end
			tooltip:Show()
		end
	elseif script == "OnLeave" then
		if self.dataObject.type == "progress bar" then
			self:SetBorderColor()
		end

		if self.dataObject.OnLeave then
			self.dataObject.OnLeave(self, ...)
		elseif not self.dataObject.OnEnter then
			tooltip:Hide()
		end
	elseif self.dataObject[script] then
		self.dataObject[script](self, ...)
	end
end

function SixDisplayMixin:RegisterCallbacks()
	local eventPrefix = "LibDataBroker_AttributeChanged_" .. self.config.name .. "_"

	LibDataBroker.UnregisterAllCallbacks(self)
	LibDataBroker.RegisterCallback(self, "LibDataBroker_DataObjectCreated", self.OnDataObjectCreated, self)

	if self.dataObject.type == "progress bar" then
		LibDataBroker.RegisterCallback(self, eventPrefix .. "value", self.OnAttributeChanged, self)
		LibDataBroker.RegisterCallback(self, eventPrefix .. "projectedValue", self.OnAttributeChanged, self)

		if not self.config.useCustomBarColor then
			LibDataBroker.RegisterCallback(self, eventPrefix .. "barR", self.OnAttributeChanged, self)
			LibDataBroker.RegisterCallback(self, eventPrefix .. "barG", self.OnAttributeChanged, self)
			LibDataBroker.RegisterCallback(self, eventPrefix .. "barB", self.OnAttributeChanged, self)
		end
	else
		if self.config.glyph then
			LibDataBroker.RegisterCallback(self, eventPrefix .. "glyph", self.OnAttributeChanged, self)
		end

		if self.config.icon then
			LibDataBroker.RegisterCallback(self, eventPrefix .. "icon", self.OnAttributeChanged, self)
			LibDataBroker.RegisterCallback(self, eventPrefix .. "iconCoords", self.OnAttributeChanged, self)
			LibDataBroker.RegisterCallback(self, eventPrefix .. "iconR", self.OnAttributeChanged, self)
			LibDataBroker.RegisterCallback(self, eventPrefix .. "iconG", self.OnAttributeChanged, self)
			LibDataBroker.RegisterCallback(self, eventPrefix .. "iconB", self.OnAttributeChanged, self)
		end

		if self.config.label then
			LibDataBroker.RegisterCallback(self, eventPrefix .. "label", self.OnAttributeChanged, self)
		end

		if self.config.text then
			LibDataBroker.RegisterCallback(self, eventPrefix .. "text", self.OnAttributeChanged, self)
		end

		if self.config.value then
			LibDataBroker.RegisterCallback(self, eventPrefix .. "value", self.OnAttributeChanged, self)
		end

		if self.config.suffix then
			LibDataBroker.RegisterCallback(self, eventPrefix .. "suffix", self.OnAttributeChanged, self)
		end
	end

	LibDataBroker.RegisterCallback(self, eventPrefix .. "visible", self.OnAttributeChanged, self)
	LibDataBroker.RegisterCallback(self, eventPrefix .. "revision", self.OnAttributeChanged, self)
end

function SixDisplayMixin:GetBarColor(alpha)
	local c = self.config.useCustomBarColor and self.config or self.dataObject
	local r, g, b, a = c.barR or 1, c.barG or 1, c.barB or 1, alpha or 1
	return r, g, b, a
end

function SixDisplayMixin:GetBarData()
	return {
		label = {current = self.cache.label or UNKNOWN, next = self.dataObject.label},
		level = {current = self.cache.level or UNKNOWN, next = self.dataObject.level},
		value = {current = self.cache.value or 0, next = tonumber(self.dataObject.value)},
	}
end

function SixDisplayMixin:GetAttributeColor(key)
	return CreateColor(self.config[key .. "R"], self.config[key .. "G"], self.config[key .. "B"])
end

function SixDisplayMixin:SetConfig(...)
	Mixin(self.config, ...)
	self.AttributeColor = {
		glyph = self:GetAttributeColor("glyph"),
		label = self:GetAttributeColor("label"),
		text = self:GetAttributeColor("text"),
		value = self:GetAttributeColor("value"),
		suffix = self:GetAttributeColor("suffix"),
	}
	self.display.glyph:SetFontObject(self.config.glyphFont or self.config.textFont)
	self.display.text:SetFontObject(self.config.textFont)
	self:SetDataObject()
	self:RegisterCallbacks()
	self:Update()
end

function SixDisplayMixin:SetBorderColor(...)
	if self.GetBorder then
		self:GetBorder():SetColor(...)
	end
end

function SixDisplayMixin:SetDataObject()
	self.dataObject = LibDataBroker:GetDataObjectByName(self.config.name) or {}
	local isProgressBar = self.dataObject.type == "progress bar"
	self.bar:SetShown(isProgressBar)
	self.projectedBar:SetShown(isProgressBar)
	self.display:SetShown(not isProgressBar)
end

function SixDisplayMixin:SetBarValue(timer)
	if self.hasTimer and not timer then
		return
	else
		self.hasTimer = false
	end

	local barData = self:GetBarData()

	if self.ShouldSkipAnimation(barData) then
		self.bar:SetValue_(barData.value.next)
		self.cache.label = barData.label.next
		self.cache.level = barData.level.next
		self.cache.value = barData.value.next
	elseif self.ShouldFillBar(barData) then
		self.bar:SetValue(100)
		self.cache.value = 100
		self.hasTimer = true
		L:ScheduleTimer(self.SetBarValue, 1, self, true)
	elseif self.ShouldResetBar(barData) then
		self.bar:SetValue_(0)
		self.cache.level = barData.level.next
		self.cache.value = 0
		self.hasTimer = true
		L:ScheduleTimer(self.SetBarValue, 0.1, self, true)
	else
		self.bar:SetValue(barData.value.next)
		self.config.value = barData.value.next
	end
end

function SixDisplayMixin:SetBarColor()
	self.bar:SetStatusBarColor(self:GetBarColor())
	self.projectedBar:SetStatusBarColor(self:GetBarColor(0.33))
end

function SixDisplayMixin:SetProjectedBarValue()
	self.projectedBar:SetValue(self.dataObject.projectedValue or 0)
end

function SixDisplayMixin:SetGlyphText()
	local glyph = ""

	if self.config.glyph then
		glyph = self.config.glyphOverride ~= "" and self.config.glyphs[self.config.glyphOverride] or self.dataObject.glyph or ""

		if glyph ~= "" then
			glyph = self.AttributeColor.glyph:WrapTextInColorCode(glyph)
		end
	end

	self.display.glyph:SetText(glyph)
	self:SetDisplayWidth()
end

function SixDisplayMixin:SetIconSizeAndTexture()
	local icon = self.config.icon and self.dataObject.icon or nil
	local size = icon and self.config.iconSize or 1

	self.display.icon:SetSize(size, size)
	self.display.icon:SetTexture(icon)
	self:SetDisplayWidth()
end

function SixDisplayMixin:SetIconTexCoords()
	local texCoords = self.dataObject.iconCoords or self.GetZoomedTexCoords(self.config.iconZoom)
	self.display.icon:SetTexCoord(unpack(texCoords))
end

function SixDisplayMixin:SetIconVertexColor()
	local r = self.dataObject.iconR or 1
	local g = self.dataObject.iconG or 1
	local b = self.dataObject.iconB or 1

	self.display.icon:SetVertexColor(r, g, b)
end

function SixDisplayMixin:SetVisibility()
	local visible = self.dataObject.visible
	local shown = (type(visible) ~= "boolean") and true or visible

	if self:IsShown() ~= shown then
		self:SetShown(shown)
		L:SendMessage("SIXDISPLAY_VISIBLITY_CHANGED", self)
	end
end

function SixDisplayMixin:SetText()
	local function getFormattedText(field, delimiter)
		if self.config[field] then
			local fieldValue = self.dataObject[field] or (field == "label" and self.config.name) or ""

			if fieldValue ~= "" then
				return self.AttributeColor[field]:WrapTextInColorCode(fieldValue .. delimiter)
			end
		end

		return ""
	end

	local formattedLabel = getFormattedText("label", ": ")
	local formattedText = getFormattedText("text", " ")
	local formattedValue = getFormattedText("value", " ")
	local formattedSuffix = getFormattedText("suffix", "")

	local combinedText = formattedLabel .. formattedText .. formattedValue .. formattedSuffix
	local textWithoutTrailingColon = string.gsub(combinedText, ":? *|r$", "|r")

	self.display.text:SetText(textWithoutTrailingColon)
	self:SetDisplayWidth()
end

function SixDisplayMixin:SetValue()
	if self.dataObject.type == "progress bar" then
		self:SetBarValue()
	end

	self:SetText()
end

function SixDisplayMixin:SetDisplayWidth()
	self.display:SetWidth(self.display.icon:GetWidth() + self.display.glyph:GetStringWidth() + self.display.text:GetStringWidth())
end

function SixDisplayMixin:Update()
	if self.dataObject.type == "progress bar" then
		self:SetBarValue()
		self:SetBarColor()
		self:SetProjectedBarValue()
	else
		self:SetGlyphText()
		self:SetIconSizeAndTexture()
		self:SetIconTexCoords()
		self:SetIconVertexColor()
		self:SetText()
	end

	self:SetVisibility()
end

SixDisplayMixin.KeyToMethod = {
	-- Bar
	projectedValue = SixDisplayMixin.SetProjectedBarValue,
	barR = SixDisplayMixin.SetBarColor,
	barG = SixDisplayMixin.SetBarColor,
	barB = SixDisplayMixin.SetBarColor,

	-- Glyph
	glyph = SixDisplayMixin.SetGlyphText,

	-- Icon
	icon = SixDisplayMixin.SetIconSizeAndTexture,
	iconCoords = SixDisplayMixin.SetIconTexCoords,
	iconR = SixDisplayMixin.SetIconVertexColor,
	iconG = SixDisplayMixin.SetIconVertexColor,
	iconB = SixDisplayMixin.SetIconVertexColor,

	-- Text
	label = SixDisplayMixin.SetText,
	text = SixDisplayMixin.SetText,
	value = SixDisplayMixin.SetValue,
	suffix = SixDisplayMixin.SetText,

	-- Visibility
	visible = SixDisplayMixin.SetVisibility,
}

-------------------------------------------------------------------------------
-- Factory
-------------------------------------------------------------------------------

local FactoryMixin = {}

function FactoryMixin.frameInitFunc(display)
	display.bar:SetMinMaxValues(0, 100)
	display.bar:SetValue(0)
	display.bar:SetOrientation("VERTICAL")
	LibSmoothStatusBar:SmoothBar(display.bar)

	display.projectedBar:SetMinMaxValues(0, 100)
	display.projectedBar:SetValue(0)
	display.projectedBar:SetOrientation("VERTICAL")
	LibSmoothStatusBar:SmoothBar(display.projectedBar)

	Mixin(display, SixDisplayMixin, {
		cache = {},
		config = CreateFromMixins(CONFIG_DEFAULTS),
	})
end

function FactoryMixin.resetterFunc(_, display)
	LibDataBroker.UnregisterAllCallbacks(display)
	table.wipe(display.cache)
	display.config = CreateFromMixins(CONFIG_DEFAULTS)
	display.bar:Hide()
	display.display:Hide()
	display.projectedBar:Hide()
	display:Hide()
	display:ClearAllPoints()
	display:SetParent(nil)
end

function FactoryMixin:GetDisplay(displayName)
	for display in self.pool:EnumerateActive() do
		if display.config.name == displayName then
			return display
		end
	end
end

function FactoryMixin:SetConfig(...)
	Mixin(self.config, ...)

	for display in self.pool:EnumerateActive() do
		display:SetConfig(self.config)
	end
end

function FactoryMixin:UpsertDisplay(displayName, ...)
	local display = self:GetDisplay(displayName) or self.pool:Acquire()

	display:SetConfig(self.config, {name = displayName}, ...)

	return display
end

-------------------------------------------------------------------------------
-- Library
-------------------------------------------------------------------------------

function L:CreateFactory(...)
	return Mixin(CreateFrame("Frame"), FactoryMixin, {
		config = CreateFromMixins(CONFIG_DEFAULTS, ...),
		pool = CreateFramePool("FRAME", UIParent, "SixDisplay_Template", FactoryMixin.resetterFunc, false, FactoryMixin.frameInitFunc),
	})
end
